#include <limits>
#include <sys/param.h>
#include <msp/core/maputils.h>
#include <msp/strings/format.h>
#include "binaryparser.h"
#include "binfloat.h"
#include "input.h"

using namespace std;

namespace Msp {
namespace DataFile {

class bad_definition: public runtime_error
{
public:
	bad_definition(const std::string &w):
		runtime_error(w)
	{ }

	virtual ~bad_definition() throw() { }
};


BinaryParser::BinaryParser(Input &i, const string &s):
	ParserMode(i, s),
	first(true),
	float_precision(32)
{
	dict[-1] = DictEntry("__kwd", "iss");
	dict[-2] = DictEntry("__str", "is");
	dict[-3] = DictEntry("__flt", "i");
}

Statement BinaryParser::parse(bool raw)
{
	while(1)
	{
		Statement st = parse_statement(raw);
		if(st.keyword=="__kwd")
		{
			int id = st.args[0].get<int>();
			if(id<=0)
				throw bad_definition("__kwd");

			const string &kw = st.args[1].get<const string &>();
			const string &args = st.args[2].get<const string &>();
			for(string::const_iterator i=args.begin(); i!=args.end(); ++i)
				for(unsigned j=0; valid_signatures[j]!=*i; ++j)
					if(!valid_signatures[j])
						throw bad_definition("__kwd");

			dict[id] = DictEntry(kw, args);
		}
		else if(st.keyword=="__str")
		{
			int id = st.args[0].get<int>();
			if(id<=0)
				throw bad_definition("__str");

			strings[id] = st.args[1].get<const string &>();
		}
		else if(st.keyword=="__flt")
			float_precision = st.args[0].get<unsigned>();
		else
			return st;

		if(raw)
			return st;
	}
}

Statement BinaryParser::parse_statement(bool raw)
{
	while(first && in.peek()=='\n')
		in.get();
	first = false;

	int id = parse_int();
	if(!in)
		return Statement();

	const DictEntry &de = get_item(dict, id);

	Statement result;
	result.keyword = de.keyword;
	result.source = src;

	for(unsigned j = 0; j<de.args.size(); ++j)
	{
		switch(de.args[j])
		{
		case IntType::signature:
			result.args.push_back(parse_int());
			break;
		case FloatType::signature:
			result.args.push_back(parse_float());
			break;
		case StringType::signature:
			result.args.push_back(parse_string());
			break;
		case BoolType::signature:
			result.args.push_back(parse_bool());
			break;
		case SymbolType::signature:
			result.args.push_back(parse_symbol());
			break;
		}
	}

	unsigned nsub = parse_int();
	for(unsigned j = 0; j<nsub; ++j)
		result.sub.push_back(parse(raw));

	result.valid = true;

	return result;
}

IntType::Store BinaryParser::parse_int()
{
	IntType::Store result = 0;
	unsigned bits = 0;

	while(in)
	{
		int c = in.get();

		result = (result<<7) | (c&0x7F);
		bits += 7;

		if(!(c&0x80))
			break;
	}

	const IntType::Store mask = 1LL<<(bits-1);
	result = (result^mask)-mask;

	return result;
}

FloatType::Store BinaryParser::parse_float()
{
	UInt64 encoded = 0;
	for(unsigned i=0; i<float_precision; i+=8)
	{
		int c = in.get();
		encoded = (encoded<<8) | (c&0xFF);
	}

	BinFloat bf = BinFloat::explode(encoded, float_precision);

	if(numeric_limits<FloatType::Store>::is_iec559)
		return bf.compose_iec559<FloatType::Store>();
	else
	{
		/* Put the float together with arithmetic since we don't know its
		internal layout */
		FloatType::Store f = 0;
		if(bf.infinity)
		{
			if(numeric_limits<FloatType::Store>::has_infinity)
				f = numeric_limits<FloatType::Store>::infinity();
			else
				f = numeric_limits<FloatType::Store>::max();
		}
		else
		{
			for(unsigned i=0; i<64; ++i)
			{
				f /= 2;
				if(bf.mantissa&1)
					f += 1;
				bf.mantissa >>= 1;
			}
			for(int i=0; i<bf.exponent; ++i)
				f *= 2;
			for(int i=0; i>bf.exponent; --i)
				f /= 2;
		}
		if(bf.sign)
			f = -f;
		return f;
	}
}

BoolType::Store BinaryParser::parse_bool()
{
	return in.get();
}

StringType::Store BinaryParser::parse_string()
{
	int len = parse_int();
	if(len>=0)
	{
		string result;
		result.reserve(len);
		for(int i = 0; i<len; ++i)
			result += in.get();
		return result;
	}
	else
		return get_item(strings, -len);
}

SymbolType::Store BinaryParser::parse_symbol()
{
	return get_item(strings, parse_int());
}

} // namespace DataFile
} // namespace Msp
